#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <arpa/inet.h>
#include <sys/stat.h>

struct request {
	uint8_t command;
	uint8_t imageid;
	uint32_t offset;
	uint32_t length;
};

enum {
	sCmdSize = 1,
	sCmdRead,
	sCmdWrite,
	sCmdEject
};

struct response {
	uint8_t status;
	uint8_t imageid;
	uint32_t length;
};

void usage(const char *progname) {
	fprintf(stderr, "Usage: %s <device>\n", progname);
	return;
}

int main(int argc, char *argv[]) {
	char *device = argv[1];
	int fd;
	int serfd;
	struct stat sb;

	if(!device) {
		usage(argv[0]);
		exit(1);
	}

	fd = open(device, O_RDWR | O_NOCTTY );
	if(fd < 0) {
		fprintf(stderr, "Could not open device %s: %d %s\n", device, errno, strerror(errno));
		exit(1);
	}

	serfd = open("serialfile", O_RDWR);
	fstat(serfd, &sb);

	struct termios opts;
	tcgetattr(fd, &opts);
	//cfsetispeed(&opts, B9600);
	//cfsetospeed(&opts, B9600);
	cfsetispeed(&opts, B57600);
	cfsetospeed(&opts, B57600);

	/* 8N1 */
	opts.c_cflag &= ~PARENB;
	opts.c_cflag &= ~CSTOPB;
	opts.c_cflag &= ~CSIZE;
	opts.c_cflag |= CS8;

	/* Disable hw flow control */
	opts.c_cflag &= ~CRTSCTS;

#if 0
	/* Software flow control */
	// opts.c_iflag |= (IXON | IXOFF | IXANY);

	/* raw input */
	//opts.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

	/* raw output */
	//opts.c_oflag &= ~OPOST;
#endif

	cfmakeraw(&opts);

	tcsetattr(fd, TCSANOW, &opts);

	do {
		struct request req;
		uint8_t *ptr = (uint8_t*)&req;
		ssize_t ret;
		ssize_t sizeread = 0;

		do {
			ret = read(fd, ptr, sizeof(req)-sizeread);
			if(ret < 0) {
				fprintf(stderr, "Error reading, read returned %zd: %d %s\n", ret, errno, strerror(errno));
			} else {
				ptr += ret;
				sizeread += ret;
				printf("Received %zd bytes of %zd\n", sizeread, sizeof(req));
			}
		} while(sizeread < sizeof(req));

		req.offset = ntohl(req.offset);
		req.length = ntohl(req.length);

		printf("Received request:\n");
		printf("Request: %x\n", req.command);
		printf("ImageID: %x\n", req.imageid);
		printf("Offset: %x\n", req.offset);
		printf("Length: %x\n", req.length);

		struct response resp;
		char *buffer = NULL;

		switch(req.command) {
		case sCmdSize:
			resp.status = 0;
			resp.imageid = req.imageid;
			resp.length = sb.st_size >> 32; // Hack, converting st_size to uint32_t
			resp.length = htonl(resp.length);
			ret = write(fd, &resp, sizeof(resp));
			printf("Wrote %zd bytes\n", ret);
			break;
		case sCmdRead:
			resp.status = 0;
			resp.imageid = req.imageid;
			resp.length = req.length;
			resp.length = htonl(resp.length);
			ret = write(fd, &resp, sizeof(resp));
			printf("Wrote %zd bytes for read cmd\n", ret);
			buffer = calloc(1, req.length);
			ret = pread(serfd, buffer, req.length, req.offset);
			printf("Read %zd bytes from file\n", ret);
			ret = write(fd, buffer, req.length);
			printf("Wrote %zd bytes from file\n", ret);
			free(buffer);
			break;
		case sCmdWrite:
			buffer = malloc(req.length);
			printf("Received write request, attempting to read %d bytes to write\n", req.length);
			sizeread = 0;
			ptr = buffer;
			do {
				ret = read(fd, ptr, req.length - sizeread);
				ptr += ret;
				sizeread += ret;
				printf("Read %zd bytes of a requested %d\n", sizeread, req.length);
			}while(sizeread < req.length);

			resp.status = 0;
			pwrite(serfd, buffer, req.length, req.offset);
			free(buffer);

			resp.imageid = req.imageid;
			resp.length = 0;
			ret = write(fd, &resp, sizeof(resp));
			printf("Wrote response size %zd\n", ret);
			break;
		case sCmdEject:
			close(fd);
			close(serfd);
			exit(0);
		};

	} while(1);
	
}
